"""
This linter checks for SELECT queries that use full table scan
"""
from collections import OrderedDict

from indexdigest.utils import explain_queries, LinterEntry, shorten_query


def check_full_table_scan(database, queries):
    """
    Full table scan

    An operation that requires reading the entire contents of a table, rather than just selected
    portions using an index. Typically performed either with small lookup tables, or in data
    warehousing situations with large tables where all available data is aggregated and analyzed.
    How frequently these operations occur, and the sizes of the tables relative to available memory,
    have implications for the algorithms used in query optimization and managing the buffer pool.

    :type database  indexdigest.database.Database
    :type queries list[str]
    :rtype: list[LinterEntry]
    """
    for (query, table_used, _, row) in explain_queries(database, queries):
        # The output from EXPLAIN shows ALL in the type column when
        # MySQL uses a full table scan to resolve a query.
        if row['type'] != 'ALL':
            continue

        context = OrderedDict()
        context['query'] = query
        context['explain_rows'] = row['rows']

        yield LinterEntry(linter_type='queries_using_full_table_scan', table_name=table_used,
                          message='"{}" query triggered full table scan'.
                          format(shorten_query(query)),
                          context=context)



"""
Database connector wrapper
"""
import logging
from collections import OrderedDict, defaultdict
from warnings import filterwarnings

import MySQLdb
from MySQLdb.cursors import DictCursor
from _mysql_exceptions import OperationalError

from indexdigest.schema import Column, Index
from indexdigest.utils import parse_dsn, memoize, IndexDigestError


class IndexDigestQueryError(IndexDigestError):
    """
    A wrapper for _mysql_exceptions.OperationalError:
    """
    pass


class DatabaseBase(object):
    """
    A generic wrapper of database connection with basic querying functionality.

    Sub-class this to mock database connection
    """

    def __init__(self, host, user, passwd, db):
        """
        Connects to a given database

        :type host str
        :type user str
        :type passwd str
        :type db str
        """
        self.logger = logging.getLogger(__name__)
        self.query_logger = logging.getLogger(__name__ + '.query')

        # lazy connect
        self._connection_params = dict(host=host, user=user, passwd=passwd, db=db)
        self._connection = None
        self.db_name = db

        # Suppress MySQL warnings when EXPLAIN is run (#63)
        filterwarnings('ignore', category=MySQLdb.Warning)

        # register queries
        self._queries = []

    @classmethod
    def connect_dsn(cls, dsn):
        """
        :type dsn str
        :rtype DatabaseBase
        """
        parsed = parse_dsn(dsn)
        return cls(**parsed)

    @property
    def connection(self):
        """
        Lazy connection

        :rtype: Connection
        """
        if self._connection is None:
            self.logger.info('Lazy connecting to %s and using %s database',
                             self._connection_params['host'], self._connection_params['db'])

            self._connection = MySQLdb.connect(**self._connection_params)

        return self._connection

    def get_queries(self):
        """
        :rtype: list[str]
        """
        return self._queries

    def query(self, sql, cursor_class=None):
        """
        :type sql str
        :type cursor_class MySQLdb.cursors.BaseCursor
        :rtype: MySQLdb.cursors.Cursor
        :raises IndexDigestQueryError
        """
        self.query_logger.info('%s', sql)

        cursor = self.connection.cursor(cursorclass=cursor_class)

        try:
            # Python 2: query should be bytes when executing %.
            # Python 3: query should be str (unicode) when executing %
            try:
                sql = sql.encode('utf8')
            except UnicodeDecodeError:
                pass

            cursor.execute(sql)
        except OperationalError as ex:
            (code, message) = ex.args  # e.g. (1054, "Unknown column 'test' in 'field list'")
            self.query_logger.error('Database error #%d: %s', code, message)
            raise IndexDigestQueryError(message)

        # register the query
        self._queries.append(sql)

        return cursor

    def query_row(self, sql):
        """
        :type sql str
        :rtype: list
        """
        return self.query(sql).fetchone()

    def query_dict_row(self, sql):
        """
        Return a single row as a dictionary

        :type sql str
        :rtype: dict
        """
        # DictCursor is a Cursor class that returns rows as dictionaries
        return self.query(sql, cursor_class=DictCursor).fetchone()

    def query_dict_rows(self, sql):
        """
        Return all rows as dictionaries

        :type sql str
        :rtype: dict[]
        """
        # DictCursor is a Cursor class that returns rows as dictionaries
        for row in self.query(sql, cursor_class=DictCursor):
            yield row

    def query_field(self, sql):
        """
        :type sql str
        :rtype: str
        """
        return self.query_row(sql)[0]

    def query_list(self, sql):
        """
        Returns an iterator with the first field on each row.

        e.g. SHOW TABLES

        :type sql str
        :rtype: list[str]
        """
        for row in self.query(sql):
            yield str(row[0])

    def query_key_value(self, sql):
        """
        Returns an ordered dictionary with key / value taken fro first two fields of each row.

        e.g. SHOW VARIABLES

        :type sql str
        :rtype: OrderedDict
        """
        res = OrderedDict()

        for row in self.query(sql):
            res[row[0]] = row[1]

        return res


class Database(DatabaseBase):
    """
    Database wrapper extended with some stats-related queries
    """

    @memoize
    def get_server_version(self):
        """
        Returns server version (e.g. "5.5.57-0+deb8u1")

        :rtype: str
        """
        return self.query_field('SELECT VERSION()')

    def get_server_hostname(self):
        """
        Return hostname of the server
        :rtype: str
        """
        return self.get_variables(like='hostname').get('hostname')

    @memoize
    def get_tables(self):
        """
        Returns the list of tables.

        :rtype: list[str]
        """
        return list(self.query_list('SHOW TABLES'))

    def get_variables(self, like=None):
        """
        Returns the key / value dictionary with server variables

        :type like str
        :rtype: OrderedDict
        """
        sql = 'SHOW VARIABLES'
        if like is not None:
            sql += ' LIKE "{}%"'.format(like)

        return self.query_key_value(sql)

    @memoize
    def explain_query(self, sql):
        """
        Runs EXPLAIN query for a given SQL

        :type sql str
        :rtype: list
        """
        # @see https://dev.mysql.com/doc/refman/5.7/en/explain-output.html
        return list(self.query_dict_rows('EXPLAIN {}'.format(sql)))

    def get_table_schema(self, table_name):
        """
        Run SHOW CREATE TABLE query for a given table
        :type table_name str
        :rtype: str
        """
        # @see https://dev.mysql.com/doc/refman/5.7/en/show-create-table.html
        return self.query_row('SHOW CREATE TABLE {}'.format(table_name))[1]

    def _get_information_schema_where(self, table_name):
        """
        :type table_name str
        :rtype: str
        """
        # @see https://dev.mysql.com/doc/refman/5.7/en/information-schema.html
        return "WHERE TABLE_SCHEMA='{db}' AND TABLE_NAME='{table_name}'".format(
            db=self._connection_params['db'], table_name=table_name)

    def get_table_metadata(self, table_name):
        """
        Return table's metadata

        :type table_name str
        :rtype: dict
        """
        # @see https://dev.mysql.com/doc/refman/5.7/en/tables-table.html
        stats = self.query_dict_row(
            "SELECT ENGINE, TABLE_ROWS, DATA_LENGTH, INDEX_LENGTH "
            "FROM information_schema.TABLES " + self._get_information_schema_where(table_name))

        return {
            'engine': stats['ENGINE'],
            'rows': stats['TABLE_ROWS'],  # For InnoDB the row count is only a rough estimate
            'data_size': stats['DATA_LENGTH'],
            'index_size': stats['INDEX_LENGTH'],
        }

    @memoize
    def get_table_columns(self, table_name):
        """
        Return the list of indices for a given table

        :type table_name str
        :rtype: list[Column]
        """
        # @see https://dev.mysql.com/doc/refman/5.7/en/show-columns.html
        columns = [
            row['Field']
            for row in self.query_dict_rows("SHOW COLUMNS FROM {}".format(table_name))
        ]

        # @see https://dev.mysql.com/doc/refman/5.7/en/columns-table.html
        rows = self.query_dict_rows(
            "SELECT COLUMN_NAME as NAME, COLUMN_TYPE as TYPE, CHARACTER_SET_NAME, COLLATION_NAME "
            "FROM information_schema.COLUMNS " + self._get_information_schema_where(table_name))

        meta = dict()

        for row in rows:
            meta[row['NAME']] = Column(name=row['NAME'], column_type=row['TYPE'],
                                       character_set=row['CHARACTER_SET_NAME'],
                                       collation=row['COLLATION_NAME'])

        # keep the order taken from SHOW COLUMNS
        return [
            meta[column]
            for column in columns
        ]

    @memoize
    def get_table_indices(self, table_name):
        """
        Return the list of indices for a given table

        :type table_name str
        :rtype: list[Index]
        """
        # @see https://dev.mysql.com/doc/refman/5.7/en/statistics-table.html
        # @see https://dev.mysql.com/doc/refman/5.7/en/show-index.html
        res = self.query_dict_rows(
            "SELECT INDEX_NAME, NON_UNIQUE, SEQ_IN_INDEX, COLUMN_NAME, CARDINALITY " +
            "FROM information_schema.STATISTICS " + self._get_information_schema_where(table_name) +
            " ORDER BY INDEX_NAME, SEQ_IN_INDEX")

        index_columns = defaultdict(list)
        index_meta = OrderedDict()

        for row in res:
            index_name = row['INDEX_NAME']
            index_columns[index_name].append(row['COLUMN_NAME'])

            if index_name not in index_meta:
                index_meta[index_name] = {
                    'unique': int(row['NON_UNIQUE']) == 0,
                    'primary': row['INDEX_NAME'] == 'PRIMARY',
                }

        ret = []

        for index_name, meta in index_meta.items():
            columns = index_columns[index_name]
            ret.append(Index(
                name=index_name, columns=columns, primary=meta['primary'], unique=meta['unique']))

        return ret



"""
Data structures for handling schema-related things like indices and columns
"""


class Index(object):
    """
    Keeps a single index meta-data
    """
    def __init__(self, name, columns, unique=False, primary=False):
        """
        :type name str
        :type columns list[str]
        :type unique bool
        :type primary bool
        """
        self._name = name
        self._columns = columns
        self._unique = unique
        self._primary = primary

    def is_covered_by(self, index):
        """
        Checks if a current index is covered by a different one

        Examples:

        PRIMARY KEY (`id`,`foo`),
        UNIQUE KEY `idx` (`id`,`foo`)  # redundant

        PRIMARY KEY (`id`),
        KEY `idx_foo` (`foo`),  # redundant (covered by idx_foo_bar)
        KEY `idx_foo_bar` (`foo`, `bar`),
        KEY `idx_id_foo` (`id`, `foo`)

        :type index Index
        :rtype: bool
        """
        # @see https://github.com/macbre/index-digest/issues/4

        # assume primary is never covered by other indices (plus self check)
        if self.is_primary or self == index:
            return False

        # equal indices - prefer unique over non unique indices
        # and primary keys over unique ones
        # @see https://github.com/macbre/index-digest/issues/49
        if self.columns == index.columns and self.is_unique:
            # we're covered by the same unique key or a primary key
            if index.is_unique or index.is_primary:
                return True

            return False

        # now take the subset of columns from the index we're comparing ourselves too
        columns_cnt = len(self.columns)

        if self.columns == index.columns[:columns_cnt]:
            if self.is_unique and index.is_primary:
                # the unique key adds a uniqueness bit to the primary key - #49
                return False

            return True

        return False

    @property
    def name(self):
        """
        :rtype: str
        """
        return self._name

    @property
    def columns(self):
        """
        :rtype: list[str]
        """
        return self._columns

    @property
    def is_unique(self):
        """
        :rtype: bool
        """
        return self._unique is True

    @property
    def is_primary(self):
        """
        :rtype: bool
        """
        return self._primary is True

    def __repr__(self):
        """
        :rtype: str
        """
        return '<{}> {}'.format(self.__class__.__name__, str(self))

    def __str__(self):
        """
        :rtype: str
        """
        return '{type}{name} ({columns})'.format(
            type='PRIMARY KEY' if self.is_primary else 'UNIQUE KEY ' if self.is_unique else 'KEY ',
            name=self.name if not self.is_primary else '',
            columns=', '.join(self.columns)
        )


class Column(object):
    """
    Keeps a single table column meta-data

    @see https://dev.mysql.com/doc/refman/5.7/en/columns-table.html
    """
    def __init__(self, name, column_type, character_set=None, collation=None):
        """
        :type name str
        :type column_type str
        :type character_set str
        :type collation str
        """
        self._name = name
        self._type = column_type
        self._character_set = character_set
        self._collation = collation

    @property
    def name(self):
        """
        :rtype: str
        """
        return self._name

    @property
    def type(self):
        """
        :rtype: str
        """
        return self._type

    @property
    def character_set(self):
        """
        :rtype: str
        """
        return self._character_set

    @property
    def collation(self):
        """
        :rtype: str
        """
        return self._collation

    def is_text_type(self):
        """
        :rtype: bool
        """
        base_type = self.type.split('(')[0].upper()
        # @see https://dev.mysql.com/doc/refman/5.7/en/string-types.html
        return base_type in \
               ['CHAR', 'VARCHAR', 'BINARY', 'VARBINARY', 'BLOB', 'TEXT', 'ENUM', 'SET']

    def is_timestamp_type(self):
        """
        :rtype: bool
        """
        base_type = self.type.upper()
        # @see https://dev.mysql.com/doc/refman/5.7/en/date-and-time-types.html
        return base_type in \
               ['DATE', 'TIME', 'DATETIME', 'TIMESTAMP', 'YEAR']

    def __repr__(self):
        """
        :rtype: str
        """
        return '<{}> {}'.format(self.__class__.__name__, str(self))

    def __str__(self):
        """
        :rtype: str
        """
        return self._name



"""
This linter checks for SELECT queries whether they trigger filesort or temporary file
"""
from collections import OrderedDict

from indexdigest.utils import explain_queries, LinterEntry, shorten_query


def filter_explain_extra(database, queries, check):
    """
    Parse "Extra" column from EXPLAIN query results, e.g.

    "Using where; Using temporary; Using filesort"

    :type database  indexdigest.database.Database
    :type queries list[str]
    :type check str
    :rtype: list
    """
    for (query, table_used, _, explain_row) in explain_queries(database, queries):
        extra_parsed = str(explain_row['Extra']).split('; ')

        if check in extra_parsed:
            context = OrderedDict()
            context['query'] = query

            context['explain_extra'] = explain_row['Extra']
            context['explain_rows'] = explain_row['rows']
            context['explain_filtered'] = explain_row.get('filtered')  # can be not set
            context['explain_key'] = explain_row['key']

            yield (query, table_used, context)


def check_queries_using_filesort(database, queries):
    """
    Using filesort

    MySQL must do an extra pass to find out how to retrieve the rows in sorted order. The sort is
    done by going through all rows according to the join type and storing the sort key and pointer
    to the row for all rows that match the WHERE clause. The keys then are sorted and the rows are
    retrieved in sorted order.

    Percona says: The truth is, filesort is badly named. Anytime a sort can't be performed from an
    index, it's a filesort. It has nothing to do with files. Filesort should be called "sort."
    It is quicksort at heart.

    :type database  indexdigest.database.Database
    :type queries list[str]
    :rtype: list[LinterEntry]
    """
    filtered = filter_explain_extra(database, queries, check='Using filesort')

    for (query, table_used, context) in filtered:
        yield LinterEntry(linter_type='queries_using_filesort', table_name=table_used,
                          message='"{}" query used filesort'.format(shorten_query(query)),
                          context=context)


def check_queries_using_temporary(database, queries):
    """
    Using temporary

    To resolve the query, MySQL needs to create a temporary table to hold the result. This
    typically happens if the query contains GROUP BY and ORDER BY clauses that list columns
    differently.

    :type database  indexdigest.database.Database
    :type queries list[str]
    :rtype: list[LinterEntry]
    """
    filtered = filter_explain_extra(database, queries, check='Using temporary')

    for (query, table_used, context) in filtered:
        yield LinterEntry(linter_type='queries_using_temporary', table_name=table_used,
                          message='"{}" query used temporary'.format(shorten_query(query)),
                          context=context)



"""
This module provides SQL query parsing functions
"""
import sqlparse

from sqlparse.sql import TokenList
from sqlparse.tokens import Name, Whitespace, Wildcard


def get_query_tokens(query):
    """
    :type query str
    :rtype: list[sqlparse.sql.Token]
    """
    tokens = TokenList(sqlparse.parse(query)[0].tokens).flatten()
    # print([(token.value, token.ttype) for token in tokens])

    return [token for token in tokens if token.ttype is not Whitespace]


def get_query_columns(query):
    """
    :type query str
    :rtype: list[str]
    """
    columns = []
    last_keyword = None
    last_token = None

    for token in get_query_tokens(query):
        if token.is_keyword and token.value.upper() not in ['AS', 'AND', 'OR']:
            # keep the name of the last keyword, e.g. SELECT, FROM, WHERE, (ORDER) BY
            last_keyword = token.value.upper()
            # print('keyword', last_keyword)
        elif token.ttype is Name:
            # analyze the name tokens, column names and where condition values
            if last_keyword in ['SELECT', 'WHERE', 'BY'] and last_token not in ['AS']:
                # print(last_keyword, last_token, token.value)

                if token.value not in columns:
                    columns.append(token.value)
        elif token.ttype is Wildcard:
            # handle wildcard in SELECT part, but ignore count(*)
            # print(last_keyword, last_token, token.value)
            if last_keyword == 'SELECT' and last_token != '(':
                columns.append(token.value)

        last_token = token.value.upper()

    return columns


def get_query_tables(query):
    """
    :type query str
    :rtype: list[str]
    """
    tables = []
    last_keyword = None
    last_token = None

    table_syntax_keywords = ['FROM', 'WHERE', 'JOIN', 'INNER JOIN', 'LEFT JOIN', 'RIGHT JOIN', 'ON']

    for token in get_query_tokens(query):
        # print([token, token.ttype])
        if token.is_keyword and token.value.upper() in table_syntax_keywords:
            # keep the name of the last keyword
            last_keyword = token.value.upper()
            # print('keyword', last_keyword)
        elif token.ttype is Name or token.is_keyword:
            # print([last_keyword, last_token, token.value])
            # analyze the name tokens, column names and where condition values
            if last_keyword in ['FROM', 'JOIN', 'INNER JOIN', 'LEFT JOIN', 'RIGHT JOIN'] \
                    and last_token not in ['AS']:
                if token.value not in tables:
                    tables.append(token.value.strip('`'))

        last_token = token.value.upper()

    return tables



# -*- coding: utf-8 -*-
"""index_digest

Analyses your database queries and schema and suggests indices improvements.

Usage:
  index_digest DSN [--sql-log=<file>]
  index_digest (-h | --help)
  index_digest --version

Options:
  DSN               Data Source Name of database to check
  --sql-log=<file>  Text file with SQL queries to check against the database
  -h --help         Show this screen.
  --version         Show version.

Examples:
  index_digest mysql://index_digest:qwerty@localhost/index_digest
  index_digest mysql://index_digest:qwerty@localhost/index_digest --sql-log=sql.log

Visit <https://github.com/macbre/index-digest>
"""
from __future__ import print_function, unicode_literals

import logging
from itertools import chain

from docopt import docopt
from termcolor import colored, cprint

import indexdigest
from indexdigest.database import Database
from indexdigest.utils import LinterEntry
from indexdigest.linters import \
    check_queries_using_filesort, check_queries_using_temporary, \
    check_not_used_indices, check_queries_not_using_indices, \
    check_not_used_tables, check_not_used_columns, \
    check_redundant_indices, \
    check_full_table_scan, \
    check_latin_columns, \
    check_selects_with_like


def format_context(context):
    """
    :type context dict
    :rtype: str
    """
    return '\n  '.join([
        "- {key}: {value}".format(
            key=colored(key, color='green', attrs=['bold']),
            value=str(value).replace("\n", "\n    ")
        )
        for (key, value) in context.items()
    ])


def main():
    """ Main entry point for CLI"""
    logger = logging.getLogger(__name__)

    arguments = docopt(__doc__, version='index_digest {}'.format(indexdigest.VERSION))
    logger.debug('Options: %s', arguments)

    if 'DSN' not in arguments:
        return

    # connect to the database
    database = Database.connect_dsn(arguments['DSN'])
    logger.debug('Connected to MySQL server v%s', database.get_server_version())

    # read SQL log file (if provided)
    sql_log = arguments.get('--sql-log')
    if sql_log:
        logger.debug('Trying to open SQL log file: %s', sql_log)

        with open(sql_log) as log_file:
            queries = log_file.readlines()
            queries = list(map(str.strip, queries))  # remove trailing spaces
            logger.debug('Got %d entries in SQL log file', len(queries))
    else:
        queries = None

    # run all checks
    reports = chain(
        check_redundant_indices(database),
        check_latin_columns(database),
    )

    # checks that use SQL log
    if queries:
        reports = chain(
            reports,
            check_not_used_indices(database, queries=queries),
            check_not_used_tables(database, queries=queries),
            check_not_used_columns(database, queries=queries),
            check_queries_not_using_indices(database, queries=queries),
            check_queries_using_filesort(database, queries=queries),
            check_queries_using_temporary(database, queries=queries),
            check_full_table_scan(database, queries=queries),
            check_selects_with_like(database, queries=queries),
        )

    # cast to a list (to be able to count reports)
    reports = list(reports)

    # emit results
    line = '-' * 60

    print(line)
    print('Found {} issue(s) to report for "{}" database'.format(len(reports), database.db_name))
    print(line)
    print('MySQL v{} at {}'.format(database.get_server_version(), database.get_server_hostname()))
    print('index-digest v{}'.format(indexdigest.VERSION))
    print(line)

    # TODO: implement formatters
    if reports:
        for report in reports:
            assert isinstance(report, LinterEntry)

            print(
                colored(report.linter_type, color='blue', attrs=['bold']) +
                ' → table affected: ' +
                colored(report.table_name, attrs=['bold'])
            )

            cprint(
                '\n{} {}'.format(colored('✗', color='red', attrs=['bold']), report.message),
                color='white')

            if report.context is not None:
                print('\n  ' + format_context(report.context))

            print()
            print(line)

        print('Queries performed: {}'.format(len(database.get_queries())))
        # print('\n'.join(map(str, database.get_queries())))
    else:
        print('Jolly, good! No issues to report')



"""
A module containing command-line tool
"""
import logging

from os import getenv

if getenv('DEBUG') == '1':
    logging.basicConfig(
        level=logging.DEBUG,
        format='%(asctime)s %(name)-35s %(levelname)-8s %(message)s',
        datefmt="%Y-%m-%d %H:%M:%S"
    )



from setuptools import setup, find_packages

from indexdigest import VERSION

# @see https://github.com/pypa/sampleproject/blob/master/setup.py
setup(
    name='indexdigest',
    version=VERSION,
    author='Maciej Brencz',
    author_email='maciej.brencz@gmail.com',
    license='MIT',
    description='Analyses your database queries and schema and suggests indices improvements',
    url='https://github.com/macbre/index-digest',
    packages=find_packages(),
    install_requires=[
        'docopt==0.6.2',
        'coverage==4.4.1',
        'pylint==1.7.4',
        'pytest==3.2.3',
        'mysqlclient==1.3.12',
        'sqlparse==0.2.4',
        'termcolor==1.1.0',
    ],
    entry_points={
        'console_scripts': [
            'index_digest=indexdigest.cli.script:main',
        ],
    }
)
